;; Generated package description from ivy-explorer.el  -*- no-byte-compile: t -*-
(define-package "ivy-explorer" "0.3.2.0.20190909.192125" "Dynamic file browsing grid using ivy" '((emacs "25") (ivy "0.10.0")) :commit "a413966cfbcecacc082d99297fa1abde0c10d3f3" :authors '(("Clemens Radermacher" . "clemera@posteo.net")) :maintainer '("Clemens Radermacher" . "clemera@posteo.net") :keywords '("convenience" "files" "matching") :url "https://github.com/clemera/ivy-explorer")
